from dataclasses import field
from typing import Optional

from pydantic.dataclasses import dataclass


@dataclass
class SnowflakeCredentials:
    """
    Credentials needed by Snowflake client
    Requires:
    - account / user / password
    Or
    - account / user / private_key
    """

    account: str
    user: str

    password: Optional[str] = field(metadata={"sensitive": True}, default=None)
    private_key: Optional[str] = field(
        metadata={"sensitive": True}, default=None
    )

    def _check_password_xor_private_key(self):
        if not self.password and not self.private_key:
            raise ValueError("Either password or private key is required")
        if self.password and self.private_key:
            raise ValueError("Can't have both private_key and password")

    def __post_init__(self):
        self._check_password_xor_private_key()



import logging
from datetime import date, datetime, timedelta
from typing import Iterator, List, Optional, Set

import requests

from ..assets import DomoAsset
from .credentials import DomoCredentials
from .endpoints import Endpoint, EndpointFactory
from .pagination import Pagination

RawData = Iterator[dict]

DOMO_PUBLIC_URL = "https://api.domo.com"
FORMAT = "%Y-%m-%d %I:%M:%S %p"
DEFAULT_TIMEOUT = 300
TOKEN_EXPIRATION_SECONDS = timedelta(seconds=3000)  # auth token lasts 1 hour

logger = logging.getLogger(__name__)


def _at_midnight(date_: date) -> datetime:
    """convert date into datetime at midnight: 00:00:00"""
    return datetime.combine(date_, datetime.min.time())


def _handle_response(response: requests.Response) -> requests.Response:
    response.raise_for_status()
    return response


def _is_expired(last_result: dict) -> bool:
    """
    Checks if the date given is expired
    i.e. date < yesterday (00:00 AM)
    """

    last_timestamp = datetime.strptime(last_result["time"], FORMAT)
    logger.info(f"Last audit input has timestamp: {last_timestamp}")
    yesterday = date.today() - timedelta(days=1)
    threshold = _at_midnight(yesterday)
    if last_timestamp < threshold:
        return True
    return False


class DomoClient:
    """
    Connect to Domo API and fetch main assets.
    https://developer.domo.com/portal/8ba9aedad3679-ap-is#platform-oauth-apis
    """

    def __init__(self, credentials: DomoCredentials):
        self._authentication = credentials.authentication
        self._bearer_headers: Optional[dict] = None
        self._token_creation_time: datetime = datetime.min
        self._endpoint_factory = EndpointFactory(credentials.base_url)
        self._private_headers = credentials.private_headers
        self._timeout = DEFAULT_TIMEOUT
        self.base_url = credentials.base_url
        self.cloud_id = credentials.cloud_id

    def _token_expired(self) -> bool:
        token_lifetime = datetime.now() - self._token_creation_time
        return token_lifetime > TOKEN_EXPIRATION_SECONDS

    def _bearer_auth(self) -> dict:
        if self._bearer_headers and not self._token_expired():
            return self._bearer_headers

        logger.info("Refreshing authentication token...")

        basic_authentication = self._authentication
        endpoint = self._endpoint_factory.authentication

        response = requests.get(
            endpoint.url(),
            auth=basic_authentication,
            timeout=self._timeout,
        )
        result = _handle_response(response).json()

        bearer_token = result["access_token"]
        self._bearer_headers = {"authorization": f"Bearer {bearer_token}"}
        self._token_creation_time = datetime.now()

        return self._bearer_headers

    def _get(
        self,
        endpoint: Endpoint,
        params: Optional[dict] = None,
        asset_id: Optional[str] = None,
    ) -> requests.Response:
        params = params if params else {}
        is_private = endpoint.is_private
        headers = self._private_headers if is_private else self._bearer_auth()
        response = requests.get(
            url=endpoint.url(asset_id),
            headers=headers,
            params=params,
            timeout=self._timeout,
        )
        return _handle_response(response)

    def _get_element(
        self,
        endpoint: Endpoint,
        params: Optional[dict] = None,
        asset_id: Optional[str] = None,
    ) -> dict:
        """Used when the response only contains one element"""
        return self._get(endpoint, params, asset_id).json()

    def _get_many(
        self,
        endpoint: Endpoint,
        params: Optional[dict] = None,
        asset_id: Optional[str] = None,
    ) -> List[dict]:
        """Used when the response contains multiple elements"""
        return self._get(endpoint, params, asset_id).json()

    def _get_paginated(self, endpoint: Endpoint) -> List[dict]:
        """Used when the response is paginated and need iterations"""
        pagination = Pagination()
        all_results: List[dict] = []
        while pagination.needs_increment:
            results = self._get_many(
                endpoint=endpoint,
                params={
                    "offset": pagination.offset,
                    "limit": pagination.per_page,
                },
            )
            all_results.extend(results)
            number_of_items = len(results)
            pagination.increment_offset(number_of_items)

            if endpoint == self._endpoint_factory.audit:
                last_result = results[-1]
                # We decided to fetch audits until the audit is prior to a particular date
                # defined in _is_expired function. To be more precise, we could have instead use
                # params start and end of the audit endpoint (timestamp in milliseconds).
                # https://developer.domo.com/portal/a4e18ca6a0c0b-retrieve-activity-log-entries
                # But we need then to manage properly the timezone conversion of the results
                # which adds more complexity than benefits
                if _is_expired(last_result):
                    pagination.should_stop = True
        return all_results

    def _datasources(self, page_id: str) -> RawData:
        endpoint = self._endpoint_factory.page_content(page_id)
        page_content = self._get_element(endpoint)
        processed: set[str] = set()
        for card in page_content["cards"]:
            for datasource in card["datasources"]:
                id_ = datasource["dataSourceId"]
                if id_ in processed:
                    continue
                yield {
                    "id": id_,
                    "name": datasource["dataSourceName"],
                    "type": datasource["dataType"],
                }
                processed.add(id_)

    def _process_pages(
        self,
        page_tree: List[dict],
        parent_path: str = "/root",
    ) -> Iterator[dict]:
        """Recursively fetch pages while building the folder architecture"""
        if not page_tree:
            return []

        for page in page_tree:
            page_id = page.get("id")
            page_name = page.get("name")
            if not (page_id and page_name):
                logger.debug(f"This page has no id nor name: {page}")
                continue

            page_children = page.get("children", list)

            detail = self._get_element(
                self._endpoint_factory.pages,
                asset_id=page_id,
            )
            datasources = self._datasources(page_id)
            yield {
                **detail,
                "datasources": list(datasources),
                "path": parent_path,
                "base_url": self.base_url,
            }
            yield from self._process_pages(
                page_tree=page_children,
                parent_path=f"{parent_path}/{page_name}",
            )

    def _link_table_to_dataset(self, dataset: dict, dataset_id: str) -> dict:
        if not self.cloud_id:
            return dataset

        lineage_endpoint = self._endpoint_factory.table_lineage(
            dataset_id=dataset_id,
            cloud_id=self.cloud_id,
        )
        dataset_table_lineage = self._get_element(lineage_endpoint)

        return {**dataset, **dataset_table_lineage}

    def _pages(self) -> RawData:
        page_hierarchy = self._get_paginated(self._endpoint_factory.pages)
        yield from self._process_pages(page_hierarchy)

    def _datasets(self) -> RawData:
        dataset_list = self._get_paginated(self._endpoint_factory.datasets)

        for dataset in dataset_list:
            dataset_id = dataset.get("id")
            if not dataset_id:
                continue

            dataset_detail = self._get_element(
                endpoint=self._endpoint_factory.datasets,
                asset_id=dataset_id,
            )

            yield {
                **self._link_table_to_dataset(
                    dataset=dataset_detail,
                    dataset_id=dataset_id,
                ),
                "base_url": self.base_url,
            }

    def _unique_datasets(self) -> RawData:
        ids_encountered: Set[str] = set()
        for dataset in self._datasets():
            dataset_id = dataset["id"]
            if dataset_id in ids_encountered:
                continue
            ids_encountered.add(dataset_id)
            yield dataset

    def _users(self) -> RawData:
        user_list = self._get_paginated(self._endpoint_factory.users)
        for user in user_list:
            user_id = user.get("id")
            if not user_id:
                continue
            yield self._get_element(
                endpoint=self._endpoint_factory.users,
                asset_id=user_id,
            )

    def _audit(self) -> RawData:
        yield from self._get_paginated(self._endpoint_factory.audit)

    def _dataflows(self) -> RawData:
        dataflows = self._get_many(self._endpoint_factory.dataflows)
        for dataflow in dataflows:
            dataflow_id = dataflow.get("id")
            if not dataflow_id:
                continue
            yield self._get_element(self._endpoint_factory.lineage(dataflow_id))

    def fetch(self, asset: DomoAsset) -> RawData:
        """Returns the needed metadata for the queried asset"""
        if asset == DomoAsset.PAGES:
            yield from self._pages()

        elif asset == DomoAsset.DATASETS:
            yield from self._unique_datasets()

        elif asset == DomoAsset.USERS:
            yield from self._users()

        elif asset == DomoAsset.AUDIT:
            yield from self._audit()

        elif asset == DomoAsset.DATAFLOWS:
            yield from self._dataflows()

        else:
            raise ValueError(f"This asset {asset} is unknown")



from enum import Enum
from typing import Dict

from .....utils import from_env


class CredentialsDbKey(Enum):
    HOST = "host"
    PORT = "port"
    DATABASE = "database"
    SCHEMA = "schema"
    USER = "user"
    PASSWORD = "password"  # noqa: S105


CREDENTIALS_ENV: Dict[CredentialsDbKey, str] = {
    CredentialsDbKey.HOST: "CASTOR_METABASE_DB_HOST",
    CredentialsDbKey.PORT: "CASTOR_METABASE_DB_PORT",
    CredentialsDbKey.DATABASE: "CASTOR_METABASE_DB_DATABASE",
    CredentialsDbKey.SCHEMA: "CASTOR_METABASE_DB_SCHEMA",
    CredentialsDbKey.USER: "CASTOR_METABASE_DB_USERNAME",
    CredentialsDbKey.PASSWORD: "CASTOR_METABASE_DB_PASSWORD",  # noqa: S105
}


def get_value(key: CredentialsDbKey, kwargs: dict) -> str:
    """
    Returns the value of the given key:
    - from kwargs in priority
    - from ENV if not provided (raises an error if not found in ENV)
    """
    env_key = CREDENTIALS_ENV[key]
    return str(kwargs.get(key.value) or from_env(env_key))


class CredentialsDb:
    """ValueObject for the credentials"""

    def __init__(
        self,
        host: str,
        port: str,
        database: str,
        schema: str,
        user: str,
        password: str,
    ):
        self.host = host
        self.port = port
        self.database = database
        self.schema = schema
        self.user = user
        self.password = password

    @classmethod
    def from_env(cls, env: dict) -> "CredentialsDb":
        """returns a new CredentialsDb with values from ENV"""
        return CredentialsDb(
            host=get_value(CredentialsDbKey.HOST, env),
            port=get_value(CredentialsDbKey.PORT, env),
            database=get_value(CredentialsDbKey.DATABASE, env),
            schema=get_value(CredentialsDbKey.SCHEMA, env),
            user=get_value(CredentialsDbKey.USER, env),
            password=get_value(CredentialsDbKey.PASSWORD, env),
        )

    def to_dict(self, hide: bool = False) -> Dict[str, str]:
        safe = (
            CredentialsDbKey.HOST,
            CredentialsDbKey.PORT,
            CredentialsDbKey.DATABASE,
            CredentialsDbKey.SCHEMA,
            CredentialsDbKey.USER,
        )
        unsafe = (CredentialsDbKey.PASSWORD,)

        def val(k: CredentialsDbKey, v: str) -> str:
            return "*" + v[-4:] if hide and k in unsafe else v

        return {a.value: val(a, getattr(self, a.value)) for a in safe + unsafe}



import json
import logging
from dataclasses import dataclass
from datetime import datetime
from enum import Enum
from typing import Literal, Optional, Set, Tuple

import requests
from dateutil.parser import parse

from .credentials import CredentialsKey, DbtCredentials, get_value

logger = logging.getLogger(__name__)

_CLOUD_URL = "https://cloud.getdbt.com/api/v2/accounts"
_DATA_KEY = "data"
_SUCCESSFUL_RUN_STATUS = 10
_DELETED_JOB_STATE = 2

Artifact = Literal["manifest.json", "run_results.json"]


class ContentType(Enum):
    JSON = "application/json"
    TEXT = "text/html"


@dataclass(frozen=True)
class DbtRun:
    id: int
    finished_at: datetime


def _is_deleted(job: dict) -> bool:
    return job["state"] == _DELETED_JOB_STATE


class DbtClient:
    """
    Connect to dbt-cloud Administrative API and downloads manifest.
    https://docs.getdbt.com/docs/dbt-cloud-apis/admin-cloud-api
    """

    def __init__(self, **kwargs):
        self._credentials = DbtCredentials(
            token=get_value(CredentialsKey.TOKEN, kwargs),
            job_id=get_value(CredentialsKey.JOB_ID, kwargs),
        )
        self._session = requests.Session()
        self._account_id: int = self._infer_account_id()

    def _headers(self, content_type: ContentType) -> dict:
        return {
            "Accept": content_type.value,
            "Authorization": "Token " + self._credentials.token,
        }

    def _call(
        self,
        url: str,
        params: Optional[dict] = None,
        content_type: ContentType = ContentType.JSON,
        ignore_not_found: bool = False,
    ) -> dict:
        headers = self._headers(content_type)
        response = self._session.get(
            url=url,
            headers=headers,
            params=params,
        )

        if ignore_not_found and response.status_code == 404:
            logger.info(f"No results found at {url} (params: {params})")
            return {}

        try:
            result = response.json()
        except:
            context = f"{url}, status {response.status_code}"
            raise ValueError(f"Couldn't extract data from {context}")
        else:
            if content_type == ContentType.JSON:
                return result[_DATA_KEY]
            return result

    def _infer_account_id(self) -> int:
        result = self._call(url=_CLOUD_URL)
        return result[0]["id"]

    def list_job_identifiers(self) -> Set[int]:
        """
        Return the IDs of all non-deleted jobs for this account
        https://docs.getdbt.com/dbt-cloud/api-v2-legacy#tag/Jobs/operation/listJobsForAccount
        """
        url = f"{_CLOUD_URL}/{self._account_id}/jobs/"
        jobs = self._call(url)
        return {job["id"] for job in jobs if not _is_deleted(job)}

    def last_run(
        self,
        job_id: Optional[int] = None,
        finished_at_range: Optional[Tuple[datetime, datetime]] = None,
    ) -> Optional[DbtRun]:
        """
        Extract the last successful run id, optionally filtered on a given datetime range
        https://docs.getdbt.com/dbt-cloud/api-v2#tag/Runs/operation/listRunsForAccount
        """
        url = f"{_CLOUD_URL}/{self._account_id}/runs/"

        params = {
            "job_definition_id": job_id or self._credentials.job_id,
            "order_by": "-finished_at",
            "status": _SUCCESSFUL_RUN_STATUS,
            "limit": 1,
        }

        if finished_at_range:
            # https://github.com/dbt-labs/dbt-cloud-openapi-spec/issues/33
            params["finished_at__range"] = json.dumps(
                [d.isoformat() for d in finished_at_range]
            )

        data = self._call(url, params)
        if not data:
            return None

        if not data[0]:
            raise ValueError("API result is empty")

        run = data[0]
        return DbtRun(id=run["id"], finished_at=parse(run["finished_at"]))

    def _fetch_run_artifact(
        self, run_id: int, artifact: Artifact, ignore_not_found=False
    ) -> Optional[dict]:
        """
        Fetch dbt manifest or run results
        https://docs.getdbt.com/dbt-cloud/api-v2-legacy#tag/Runs/operation/getArtifactsByRunId
        """
        url = f"{_CLOUD_URL}/{self._account_id}/runs/{run_id}/artifacts/{artifact}"
        logger.info(
            f"Extracting {artifact} from run id {run_id} with url {url}"
        )

        # setting text content as a workaround to this issue
        # https://stackoverflow.com/questions/68201659/dbt-cloud-api-to-extract-run-artifacts
        return self._call(
            url,
            content_type=ContentType.TEXT,
            ignore_not_found=ignore_not_found,
        )

    def fetch_manifest(self, run_id: Optional[int] = None) -> dict:
        """
        Fetch the manifest or the given run, or of the latest run for the configured job
        """
        if not run_id:
            run = self.last_run()
            assert run, f"No run found for job {self._credentials.job_id}"
            run_id = run.id
        manifest = self._fetch_run_artifact(run_id, "manifest.json")
        assert manifest, f"No manifest found for run id {run_id}"
        return manifest

    def fetch_run_results(self, run_id: int) -> Optional[dict]:
        """
        Fetch the run results of the given run.
        It seems the API doesn't always find the file, in which case None is returned.
        """
        return self._fetch_run_artifact(
            run_id, "run_results.json", ignore_not_found=True
        )



from typing import Iterable, Set, Tuple

from .sdk import LookmlModel


def lookml_explore_names(
    lookmls: Iterable[LookmlModel],
) -> Set[Tuple[str, str]]:
    """
    Explores from the lookml models
    Only valid explores are yielded: with all infos
    """
    model_explores = (
        (model, explore)
        for model in lookmls
        for explore in model.explores or []
    )

    return {
        (model.name, explore.name)
        for model, explore in model_explores
        # accept hidden resources
        if model.name and explore.name
    }



import logging

from ...utils import LocalStorage, from_env, write_summary
from ..abstract import (
    CATALOG_ASSETS,
    EXTERNAL_LINEAGE_ASSETS,
    QUERIES_ASSETS,
    VIEWS_ASSETS,
    ExtractionProcessor,
    SupportedAssets,
    WarehouseAsset,
    WarehouseAssetGroup,
    common_args,
    extractable_asset_groups,
)
from .client import RedshiftClient
from .query import RedshiftQueryBuilder

logger = logging.getLogger(__name__)

REDSHIFT_ASSETS: SupportedAssets = {
    WarehouseAssetGroup.CATALOG: CATALOG_ASSETS,
    WarehouseAssetGroup.QUERY: QUERIES_ASSETS,
    WarehouseAssetGroup.VIEW_DDL: VIEWS_ASSETS,
    WarehouseAssetGroup.ROLE: (
        WarehouseAsset.USER,
        WarehouseAsset.GROUP,
    ),
    WarehouseAssetGroup.EXTERNAL_LINEAGE: EXTERNAL_LINEAGE_ASSETS,
}

REDSHIFT_USER = "CASTOR_REDSHIFT_USER"
REDSHIFT_PASSWORD = "CASTOR_REDSHIFT_PASSWORD"  # noqa: S105
REDSHIFT_HOST = "CASTOR_REDSHIFT_HOST"
REDSHIFT_PORT = "CASTOR_REDSHIFT_PORT"
REDSHIFT_DATABASE = "CASTOR_REDSHIFT_DATABASE"


def _credentials(params: dict) -> dict:
    """extract Redshift credentials"""

    return {
        "user": params.get("user") or from_env(REDSHIFT_USER),
        "password": params.get("password") or from_env(REDSHIFT_PASSWORD),
        "host": params.get("host") or from_env(REDSHIFT_HOST),
        "port": params.get("port") or from_env(REDSHIFT_PORT),
        "database": params.get("database") or from_env(REDSHIFT_DATABASE),
    }


def extract_all(**kwargs) -> None:
    """
    Extract all assets from Redshift and store the results in CSV files
    """
    output_directory, skip_existing = common_args(kwargs)

    client = RedshiftClient(credentials=_credentials(kwargs))

    query_builder = RedshiftQueryBuilder()

    storage = LocalStorage(directory=output_directory)

    extractor = ExtractionProcessor(
        client=client,
        query_builder=query_builder,
        storage=storage,
    )

    for group in extractable_asset_groups(REDSHIFT_ASSETS):
        for asset in group:
            logger.info(f"Extracting `{asset.value.upper()}` ...")
            location = extractor.extract(asset, skip_existing)
            logger.info(f"Results stored to {location}\n")

    write_summary(
        output_directory,
        storage.stored_at_ts,
        client_name=client.name(),
    )



from ...types import ExternalAsset


class PowerBiAsset(ExternalAsset):
    """PowerBi assets"""

    ACTIVITY_EVENTS = "activity_events"
    DASHBOARDS = "dashboards"
    DATASETS = "datasets"
    DATASET_FIELDS = "dataset_fields"
    METADATA = "metadata"
    REPORTS = "reports"
    TABLES = "tables"
    USERS = "users"


# Assets extracted from the Metadata file
# They are not directly fetched from the PowerBi api.
METADATA_ASSETS = (
    PowerBiAsset.DATASET_FIELDS,
    PowerBiAsset.TABLES,
    PowerBiAsset.USERS,
)



import pytest

from . import SafeMode, safe_mode


class CustomException(BaseException):
    pass


def test_no_safe_mode_no_exception():
    @safe_mode()
    def foo():
        return "foo"

    assert foo() == "foo"


def test_no_safe_mode_exception():
    @safe_mode(None)
    def foo():
        raise CustomException()

    with pytest.raises(CustomException):
        foo()


def test_safe_mode_no_exception():
    safe_params = SafeMode((CustomException,), 1)

    @safe_mode(safe_params)
    def foo():
        return "foo"

    assert foo() == "foo"


def test_safe_mode_exception():
    safe_params = SafeMode((CustomException,), 1)

    @safe_mode(safe_params)
    def foo():
        raise CustomException()

    assert foo() is None


def test_safe_mode_other_exception():
    safe_params = SafeMode((CustomException,), 1)

    @safe_mode(safe_params)
    def foo():
        raise Exception()

    with pytest.raises(Exception):
        foo()


def test_safe_mode_exception_default():
    safe_params = SafeMode((CustomException,), 1)

    @safe_mode(safe_params, lambda: "default")
    def foo():
        raise CustomException()

    assert foo() == "default"


def test_safe_mode_max_exception():
    safe_params = SafeMode((CustomException,), 2)

    @safe_mode(safe_params, lambda: "default")
    def foo():
        raise CustomException()

    assert foo() == "default"
    assert foo() == "default"
    with pytest.raises(CustomException):
        foo()
    assert len(safe_params.errors_caught) == 2


def test_safe_mode_max_no_decorator():
    safe_params = SafeMode((CustomException,), 2)

    def foo(arg1):
        return arg1

    def bar():
        return safe_mode(safe_params)(foo)("bar")

    assert bar() == "bar"


def test_safe_mode_max_no_decorator_exception():
    safe_params = SafeMode((CustomException,), 2)

    def foo():
        raise CustomException()

    def bar():
        default = lambda: "default"
        return safe_mode(safe_params, default)(foo)()

    assert bar() == "default"
    assert bar() == "default"
    with pytest.raises(CustomException):
        bar()
    assert len(safe_params.errors_caught) == 2



from typing import Iterator

import pymssql  # type: ignore
from sqlalchemy import text

from ...utils import ExtractionQuery, SqlalchemyClient, uri_encode

SERVER_URI = "{user}:{password}@{host}:{port}/{database}"
MSSQL_URI = f"mssql+pymssql://{SERVER_URI}"
DEFAULT_PORT = 1433

_KEYS = ("user", "password", "host", "port", "database")


def _check_key(credentials: dict) -> None:
    for key in credentials:
        if key not in _KEYS:
            raise KeyError(f"Missing {key} in credentials")


class MSSQLClient(SqlalchemyClient):
    """Microsoft Server SQL client"""

    @staticmethod
    def name() -> str:
        return "MSSQL"

    def _engine_options(self, credentials: dict) -> dict:
        return {}

    def _build_uri(self, credentials: dict) -> str:
        _check_key(credentials)
        uri = MSSQL_URI.format(
            user=credentials["user"],
            password=uri_encode(credentials["password"]),
            host=credentials["host"],
            port=credentials.get("port") or DEFAULT_PORT,
            database=credentials["database"],
        )
        return uri

    def execute(self, query: ExtractionQuery) -> Iterator[dict]:
        """
        Re-implements the SQLAlchemyClient execute function to ensure we consume
        the cursor before calling connection.close() as it wipes out the data
        otherwise
        """
        connection = self.connect()
        try:
            proxy = connection.execute(text(query.statement), query.params)
            results = list(self._process_result(proxy))
            yield from results
        finally:
            self.close()



import os
import unittest
from unittest import mock

import requests_mock  # type: ignore
import tableauserverclient as TSC  # type: ignore

from ....client import ApiClient
from ..utils import KEYS

TEST_ASSET_DIR = os.path.join(os.path.dirname(__file__), "../assets/rest_api/")

GET_XML = os.path.join(TEST_ASSET_DIR, "project_get.xml")

EXPECTED = [
    {
        "id": "ee8c6e70-43b6-11e6-af4f-f7b0d8e20760",
        "name": "default",
        "description": "The default project that was automatically created by Tableau.",
        "parent_id": None,
        "folder_path": "/default",
    },
    {
        "id": "1d0304cd-3796-429f-b815-7258370b9b74",
        "name": "Tableau",
        "description": None,
        "parent_id": None,
        "folder_path": "/Tableau",
    },
    {
        "id": "4cc52973-5e3a-4d1f-a4fb-5b5f73796edf",
        "name": "Tableau/Child_1",
        "description": None,
        "parent_id": "1d0304cd-3796-429f-b815-7258370b9b74",
        "folder_path": "/Tableau/Tableau/Child_1",
    },
]


class ProjectTests(unittest.TestCase):
    @mock.patch.dict(os.environ, KEYS)
    def setUp(self):
        self._client = ApiClient()
        self._client._server = TSC.Server("http://test")

        # Fake signin
        self._client._server._site_id = "dad65087-b08b-4603-af4e-2887b8aafc67"
        self._client._server._auth_token = "j80k54ll2lfMZ0tv97mlPvvSCRyD0DOM"

        self.baseurl = self._client._server.projects.baseurl

    def test_fetch_projects(self):
        with open(GET_XML, "rb") as f:
            response_xml = f.read().decode("utf-8")
        with requests_mock.mock() as m:
            m.get(self.baseurl, text=response_xml)
            results = self._client._fetch_projects()

        self.assertEqual(results, EXPECTED)



import logging
from typing import List, Optional, Set, Tuple

from google.api_core.exceptions import Forbidden
from google.api_core.page_iterator import Iterator as PageIterator
from google.cloud.bigquery import Client as GoogleCloudClient  # type: ignore
from google.cloud.bigquery.dataset import Dataset  # type: ignore
from google.oauth2.service_account import Credentials  # type: ignore

from ...utils import SqlalchemyClient, retry

logger = logging.getLogger(__name__)

BIGQUERY_URI = "bigquery://"

CREDENTIALS_INFO_KEY = "credentials_info"
PROJECT_ID_KEY = "project_id"

_RETRY_NUMBER = 1
_RETRY_BASE_MS = 60_000


class BigQueryClient(SqlalchemyClient):
    """Connect to BigQuery and run SQL queries"""

    def __init__(
        self,
        credentials: dict,
        db_allowed: Optional[Set[str]] = None,
        db_blocked: Optional[Set[str]] = None,
        dataset_blocked: Optional[Set[str]] = None,
    ):
        super().__init__(credentials)
        self._db_allowed = db_allowed
        self._db_blocked = db_blocked
        self._dataset_blocked = dataset_blocked
        self.client = self._client()
        self._projects: List[str] | None = None
        self._datasets: List[Dataset] | None = None

    @staticmethod
    def name() -> str:
        return "BigQuery"

    def _keep_project(self, project: str) -> bool:
        if self._db_allowed and project not in self._db_allowed:
            return False
        if self._db_blocked and project in self._db_blocked:
            return False
        return True

    def _keep_dataset(self, dataset: str) -> bool:
        if not self._dataset_blocked:
            return True

        return dataset not in self._dataset_blocked

    def _engine_options(self, credentials: dict) -> dict:
        return {
            CREDENTIALS_INFO_KEY: credentials,
        }

    def _build_uri(self, credentials: dict) -> str:
        return BIGQUERY_URI

    def _client(self) -> GoogleCloudClient:
        assert (
            CREDENTIALS_INFO_KEY in self._options
        ), "Missing BigQuery credentials in engine's options"
        credentials = self._options[CREDENTIALS_INFO_KEY]
        return GoogleCloudClient(
            project=credentials.get(PROJECT_ID_KEY),
            credentials=Credentials.from_service_account_info(credentials),
        )

    def _list_datasets(self) -> List[Dataset]:
        """
        Returns datasets available for the given GCP client
        Cache the result in self._datasets to reduce number of API calls
        """
        if self._datasets is None:
            self._datasets = [
                dataset
                for project_id in self.get_projects()
                for dataset in self.client.list_datasets(project_id)
                if self._keep_dataset(dataset.dataset_id)
            ]
        return self._datasets

    @retry(
        (Forbidden,),
        max_retries=_RETRY_NUMBER,
        base_ms=_RETRY_BASE_MS,
        log_exc_info=True,
    )
    def get_projects(self) -> List[str]:
        """
        Returns distinct project_id available for the given GCP client
        Cache the result in self._projects to reduce number of API calls.

        Note: Calling list_projects from GoogleCloudClient causes some
        ```
        google.api_core.exceptions.Forbidden: 403 GET https://bigquery.googleapis.com/bigquery/v2/projects?prettyPrint=false
        Quota exceeded: Your user exceeded quota for concurrent project.lists requests.
        ````
        In that case, task should retry automatically.
        """
        if self._projects is None:
            self._projects = [
                p.project_id
                for p in self.client.list_projects(retry=None)  # type: ignore
                if self._keep_project(p.project_id)
            ]
        return self._projects

    def get_regions(self) -> Set[Tuple[str, str]]:
        """
        Returns distinct (project_id, region) available for the given GCP client
        """
        return {
            (ds.project, ds._properties["location"])
            for ds in self._list_datasets()
        }

    def get_datasets(self) -> Set[Tuple[str, str]]:
        """
        Returns distinct (project_id, dataset_id) available for the given GCP client
        """
        return {(ds.project, ds.dataset_id) for ds in self._list_datasets()}



# Fields which will be use for Tableau GraphQL API
from enum import Enum
from typing import Dict, List, Union

from .assets import TableauAsset, TableauGraphqlAsset

FIELDS = "fields"
OBJECT_TYPE = "object_type"


class GQLQueryFields(Enum):
    BIN_FIELDS: str = """
           datasource {
               ... on PublishedDatasource {
                    name
                    luid
               }
           }
           dataType
           description
           fields {
               id
           }
           folderName
           id
           name
           role
    """

    CALCULATED_FIELDS: str = """
           datasource {
               ... on PublishedDatasource {
                    name
                    luid
               }
           }
           fields {
               id
           }
           dataType
           description
           folderName
           id
           name
           role

    """

    COLUMN_FIELDS: str = """
           columns {
               id
               name
               table {
                   ... on DatabaseTable {
                       id
                       description
                       name
                       fullName
                       schema
                       database {
                           name
                           luid
                       }
                   }
               }

           }
           datasource {
               ... on PublishedDatasource {
                    name
                    luid
               }
           }
           dataType
           description
           folderName
           id
           name
           role
    """

    CUSTOM_SQL_TABLE: str = """
           id
           name
           columns {
               referencedByFields {
                   datasource {
                       ... on PublishedDatasource {
                           luid
                       }

                       ... on EmbeddedDatasource {
                           id
                       }
                   }
               }
           }
    """

    CUSTOM_SQL_QUERY: str = """
           id
           name
           query
           database {
               name
               connectionType
           }
           tables {
               name
           }
    """

    DASHBOARDS: str = """
    id
    name
    path
    tags {
        name
    }
    workbook {
       luid # to retrieve the parent
    }
    """

    DATASOURCE: str = """
           ... on PublishedDatasource {
               luid
           }
           id
           name
           hasExtracts
           upstreamTables {
               id
               schema
               name
               fullName
               database {
                   id
                   name
                   connectionType
               }
           }
    """

    GROUP_FIELDS: str = """
           datasource {
               ... on PublishedDatasource {
                    name
                    luid
               }
           }
           dataType
           description
           fields {
               id
           }
           folderName
           id
           name
           role
    """

    WORKBOOK_TO_DATASOURCE: str = """
           luid
           id
           embeddedDatasources {
               id
               name
           }
           upstreamDatasources {
               luid
               name
           }
    """


QueryInfo = List[Dict[str, Union[GQLQueryFields, TableauGraphqlAsset]]]

QUERY_FIELDS: Dict[TableauAsset, QueryInfo] = {
    TableauAsset.CUSTOM_SQL_TABLE: [
        {
            FIELDS: GQLQueryFields.CUSTOM_SQL_TABLE,
            OBJECT_TYPE: TableauGraphqlAsset.CUSTOM_SQL,
        },
    ],
    TableauAsset.CUSTOM_SQL_QUERY: [
        {
            FIELDS: GQLQueryFields.CUSTOM_SQL_QUERY,
            OBJECT_TYPE: TableauGraphqlAsset.CUSTOM_SQL,
        },
    ],
    TableauAsset.DASHBOARD: [
        {
            FIELDS: GQLQueryFields.DASHBOARDS,
            OBJECT_TYPE: TableauGraphqlAsset.DASHBOARD,
        },
    ],
    TableauAsset.DATASOURCE: [
        {
            FIELDS: GQLQueryFields.DATASOURCE,
            OBJECT_TYPE: TableauGraphqlAsset.DATASOURCE,
        },
    ],
    TableauAsset.FIELD: [
        {
            FIELDS: GQLQueryFields.BIN_FIELDS,
            OBJECT_TYPE: TableauGraphqlAsset.BIN_FIELD,
        },
        {
            FIELDS: GQLQueryFields.CALCULATED_FIELDS,
            OBJECT_TYPE: TableauGraphqlAsset.CALCULATED_FIELD,
        },
        {
            FIELDS: GQLQueryFields.COLUMN_FIELDS,
            OBJECT_TYPE: TableauGraphqlAsset.COLUMN_FIELD,
        },
        {
            FIELDS: GQLQueryFields.GROUP_FIELDS,
            OBJECT_TYPE: TableauGraphqlAsset.GROUP_FIELD,
        },
    ],
    TableauAsset.WORKBOOK_TO_DATASOURCE: [
        {
            FIELDS: GQLQueryFields.WORKBOOK_TO_DATASOURCE,
            OBJECT_TYPE: TableauGraphqlAsset.WORKBOOK_TO_DATASOURCE,
        },
    ],
}



from argparse import ArgumentParser

from castor_extractor.visualization import looker  # type: ignore


def main():
    parser = ArgumentParser()
    parser.add_argument("-b", "--base-url", help="Looker base url")
    parser.add_argument("-u", "--username", help="Looker client id")
    parser.add_argument("-p", "--password", help="Looker client secret")
    parser.add_argument("-o", "--output", help="Directory to write to")
    parser.add_argument("-t", "--timeout", type=int, help="Timeout in seconds")
    parser.add_argument(
        "--thread-pool-size",
        type=int,
        help="Thread pool size, if searching per folder",
    )
    parser.add_argument(
        "--safe-mode",
        "-s",
        help="Looker safe mode",
        action="store_true",
    )
    parser.add_argument(
        "--log-to-stdout",
        help="Send all log outputs to stdout instead of stderr",
        action="store_true",
    )

    parser.add_argument(
        "--search-per-folder",
        help="Fetches Looks and Dashboards per folder",
        action="store_true",
    )

    args = parser.parse_args()

    looker.extract_all(
        base_url=args.base_url,
        client_id=args.username,
        client_secret=args.password,
        log_to_stdout=args.log_to_stdout,
        output_directory=args.output,
        safe_mode=args.safe_mode,
        search_per_folder=args.search_per_folder,
        thread_pool_size=args.thread_pool_size,
        timeout=args.timeout,
    )



import enum
import json
from datetime import date, datetime

import pytest

from .object import deep_serialize


class _User:
    """dictable object"""

    def __init__(self, name: str):
        self.name = name

    def __iter__(self):
        for key in self.__dict__:
            yield key, getattr(self, key)


def test_deep_serialize__None():
    assert deep_serialize(None) == None


def test_deep_serialize__str():
    assert deep_serialize("a") == "a"


def test_deep_serialize__Enum():
    class MyEnum(enum.Enum):
        HELLO = "world"

    assert deep_serialize(MyEnum.HELLO) == "world"


def test_deep_serialize__int():
    assert deep_serialize(1) == 1
    assert deep_serialize(12345678987654321) == 12345678987654321


def test_deep_serialize__float():
    assert deep_serialize(1.3) == 1.3
    assert deep_serialize(0.01) == 0.01


def test_deep_serialize__date():
    assert deep_serialize(date(2020, 1, 2)) == "2020-01-02"


def test_deep_serialize__datetime():
    dt = datetime(1989, 8, 6, 10, 30)
    assert deep_serialize(dt) == "1989-08-06T10:30:00"


def test_deep_serialize__tuple():
    assert deep_serialize(tuple()) == []
    assert deep_serialize((1, "4")) == [1, "4"]


def test_deep_serialize__list():
    assert deep_serialize([]) == []
    assert deep_serialize([1, 2]) == [1, 2]
    assert deep_serialize([1.3, "4.3"]) == [1.3, "4.3"]


def test_deep_serialize__dict():
    assert deep_serialize({}) == {}
    assert deep_serialize({"a": 1}) == {"a": 1}


def test_deep_serialize__complex_object():
    user = _User(name="OK")
    serialized = deep_serialize(user)
    assert isinstance(serialized, dict)
    assert serialized["name"] == "OK"

    # Object non-dictable
    class Test:
        attr = 1

    with pytest.raises(ValueError):
        deep_serialize(Test())


def test_deep_serialize__nested():
    user1 = _User(name="1")
    user2 = _User(name="2")
    some_date = date(2021, 7, 1)

    input = [
        {
            "users": [user1, user2],
            "extracted_at": None,
            "level1": {"levels": [{2: 1}, user1]},
        },
        {"users": [], "extracted_at": some_date},
    ]

    serialized = deep_serialize(input)

    assert len(serialized) == 2

    d1, d2 = serialized

    d1_user_names = list(map(lambda x: x["name"], d1["users"]))
    assert d1_user_names == ["1", "2"]
    assert d1["level1"]["levels"][0] == {2: 1}
    assert d2["extracted_at"] == "2021-07-01"

    assert len(json.dumps(serialized)) > 1



from ...types import ExternalAsset


class MetabaseAsset(ExternalAsset):
    """Metabase assets"""

    COLLECTION = "collection"
    USER = "user"
    DATABASE = "database"
    TABLE = "table"
    CARD = "card"
    DASHBOARD = "dashboard"
    DASHBOARD_CARDS = "dashboard_cards"


EXPORTED_FIELDS = {
    MetabaseAsset.COLLECTION: (
        "id",
        "name",
        "description",
        "color",
        "archived",
        "location",
        "personal_owner_id",
        "slug",
        "namespace",
    ),
    MetabaseAsset.USER: (
        "id",
        "email",
        "first_name",
        "last_name",
        "date_joined",
        "last_login",
        "is_superuser",
        "is_active",
        "updated_at",
        "locale",
    ),
    MetabaseAsset.DATABASE: (
        "id",
        "created_at",
        "updated_at",
        "name",
        "description",
        "engine",
        "is_sample",
        "is_full_sync",
        "points_of_interest",
        "caveats",
        "metadata_sync_schedule",
        "cache_field_values_schedule",
        "timezone",
        "is_on_demand",
        "options",
        "auto_run_queries",
        "dbname",  # calculated field
    ),
    MetabaseAsset.TABLE: (
        "id",
        "created_at",
        "updated_at",
        "name",
        "description",
        "entity_name",
        "entity_type",
        "active",
        "db_id",
        "display_name",
        "visibility_type",
        "schema",
        "points_of_interest",
        "caveats",
        "show_in_getting_started",
        "field_order",
    ),
    MetabaseAsset.CARD: (
        "id",
        "created_at",
        "updated_at",
        "name",
        "description",
        "display",
        "dataset_query",
        "visualization_settings",
        "creator_id",
        "database_id",
        "table_id",
        "query_type",
        "archived",
        "collection_id",
        "made_public_by_id",
        "enable_embedding",
        "embedding_params",
        "result_metadata",
        "collection_position",
        "view_count",
    ),
    MetabaseAsset.DASHBOARD: (
        "id",
        "created_at",
        "updated_at",
        "name",
        "description",
        "creator_id",
        "parameters",
        "points_of_interest",
        "caveats",
        "show_in_getting_started",
        "made_public_by_id",
        "enable_embedding",
        "embedding_params",
        "archived",
        "position",
        "collection_id",
        "collection_position",
        "view_count",
    ),
    MetabaseAsset.DASHBOARD_CARDS: (
        "id",
        "card_id",
        "dashboard_id",
        "created_at",
        "updated_at",
        "sizex",
        "sizey",
        "row",
        "col",
        "parameter_mappings",
        "visualization_settings",
    ),
}



import logging
from argparse import ArgumentParser

from castor_extractor.visualization import domo  # type: ignore

logging.basicConfig(level=logging.INFO, format="%(levelname)s - %(message)s")


def main():
    parser = ArgumentParser()

    # mandatory
    parser.add_argument(
        "-b",
        "--base-url",
        help="Host name where the server is running",
    )
    parser.add_argument(
        "-a",
        "--api-token",
        help="Token to connect to Domo",
    )
    parser.add_argument(
        "-d",
        "--developer-token",
        help="Developer token to use private API",
    )
    parser.add_argument(
        "-c",
        "--client-id",
        help="Client id of the Domo account",
    )
    parser.add_argument(
        "-C",
        "--cloud-id",
        help="Cloud id is the external warehouse id within Domo",
    )

    parser.add_argument("-o", "--output", help="Directory to write to")

    args = parser.parse_args()

    domo.extract_all(
        api_token=args.api_token,
        base_url=args.base_url,
        client_id=args.client_id,
        cloud_id=args.cloud_id,
        developer_token=args.developer_token,
        output_directory=args.output,
    )



from typing import Dict


class SalesforceCredentials:
    """
    Class to handle Salesforce rest API permissions
    """

    def __init__(
        self,
        *,
        username: str,
        password: str,
        security_token: str,
        consumer_key: str,
        consumer_secret: str,
    ):
        self.username = username
        self.password = password + security_token
        self.consumer_key = consumer_key
        self.consumer_secret = consumer_secret

    def token_request_payload(self) -> Dict[str, str]:
        """
        Params to post to the API in order to retrieve the authentication token
        """
        return {
            "grant_type": "password",
            "client_id": self.consumer_key,
            "client_secret": self.consumer_secret,
            "username": self.username,
            "password": self.password,
        }



import logging

from ..abstract import (
    CATALOG_ASSETS,
    EXTERNAL_LINEAGE_ASSETS,
    QUERIES_ASSETS,
    VIEWS_ASSETS,
    SupportedAssets,
    WarehouseAsset,
    WarehouseAssetGroup,
)

logger = logging.getLogger(__name__)

SYNAPSE_ASSETS: SupportedAssets = {
    WarehouseAssetGroup.CATALOG: CATALOG_ASSETS,
    WarehouseAssetGroup.QUERY: QUERIES_ASSETS,
    WarehouseAssetGroup.VIEW_DDL: VIEWS_ASSETS,
    WarehouseAssetGroup.ROLE: (WarehouseAsset.USER,),
    WarehouseAssetGroup.EXTERNAL_LINEAGE: EXTERNAL_LINEAGE_ASSETS,
}



from ..assets import SalesforceReportingAsset

queries = {
    SalesforceReportingAsset.DASHBOARDS: """
        SELECT
            CreatedBy.Id,
            CreatedDate,
            Description,
            DeveloperName,
            FolderId,
            FolderName,
            IsDeleted,
            LastReferencedDate,
            LastViewedDate,
            NamespacePrefix,
            RunningUserId,
            Title,
            Type
        FROM Dashboard
    """,
    SalesforceReportingAsset.REPORTS: """
        SELECT
            Description,
            DeveloperName,
            FolderName,
            Format,
            IsDeleted,
            LastReferencedDate,
            LastRunDate,
            LastViewedDate,
            Name,
            NamespacePrefix,
            OwnerId
        FROM Report
    """,
    SalesforceReportingAsset.USERS: """
        SELECT
            Id,
            Email,
            FirstName,
            LastName,
            CreatedDate
        FROM User
    """,
}



from typing import Dict, Tuple

from ...types import ExternalAsset


class QlikAsset(ExternalAsset):
    """Qlik assets"""

    SPACES = "spaces"
    USERS = "users"
    APPS = "apps"
    LINEAGE = "lineage"
    MEASURES = "measures"
    CONNECTIONS = "connections"


EXPORTED_FIELDS: Dict[QlikAsset, Tuple[str, ...]] = {
    QlikAsset.SPACES: (
        "id",
        "type",
        "ownerId",
        "tenantId",
        "name",
        "description",
        "meta",
        "links",
        "createdAt",
        "createdBy",
        "updatedAt",
    ),
    QlikAsset.USERS: (
        "id",
        "tenantId",
        "status",
        "name",
        "email",
        "links",
    ),
    QlikAsset.APPS: (
        "name",
        "spaceId",
        "description",
        "thumbnailId",
        "resourceAttributes",
        "resourceCustomAttributes",
        "resourceUpdatedAt",
        "resourceType",
        "resourceId",
        "resourceCreatedAt",
        "id",
        "createdAt",
        "updatedAt",
        "creatorId",
        "updaterId",
        "tenantId",
        "isFavorited",
        "links",
        "actions",
        "collectionIds",
        "meta",
        "ownerId",
        "resourceReloadEndTime",
        "resourceReloadStatus",
        "resourceSize",
        "itemViews",
    ),
    QlikAsset.MEASURES: (
        "qInfo",
        "qMeta",
        "qData",
    ),
    QlikAsset.CONNECTIONS: (
        "id",
        "links",
        "privileges",
        "qArchitecture",
        "qConnectStatement",
        "qEngineObjectID",
        "qID",
        "qLogOn",
        "qName",
        "qType",
        "tenant",
    ),
}



from ...types import ExternalAsset


class ModeAnalyticsAsset(ExternalAsset):
    """Mode Analytics assets"""

    DATASOURCE = "data_sources"
    REPORT = "reports"
    COLLECTION = "spaces"  # legacy name, still valid in the API
    MEMBER = "user"
    QUERY = "queries"


ASSETS_WITH_OWNER = (
    ModeAnalyticsAsset.COLLECTION,
    ModeAnalyticsAsset.REPORT,
)


EXPORTED_FIELDS = {
    ModeAnalyticsAsset.DATASOURCE: (
        "id",
        "token",
        "name",
        "host",
        "display_name",
        "description",
        "database",
        "provider",
        "vendor",
        "adapter",
        "created_at",
        "public",
        "queryable",
    ),
    ModeAnalyticsAsset.MEMBER: (
        "id",
        "token",
        "username",
        "name",
        "email",
        "created_at",
    ),
    ModeAnalyticsAsset.REPORT: (
        "id",
        "token",
        "name",
        "description",
        "type",
        "archived",
        "space_token",
        "public",
        "created_at",
        "updated_at",
        "published_at",
        "edited_at",
        "is_embedded",
        "query_count",
        "chart_count",
        "view_count",
        "creator",
    ),
    ModeAnalyticsAsset.QUERY: (
        "id",
        "token",
        "name",
        "data_source_id",
        "raw_query",
        "report_token",
    ),
    ModeAnalyticsAsset.COLLECTION: (
        "id",
        "token",
        "space_type",
        "name",
        "description",
        "state",
        "default_access_level",
        "creator",
    ),
}



from collections import namedtuple
from unittest.mock import patch

from castor_extractor.visualization.looker.api.sdk import (  # type: ignore
    has_admin_permissions,
)
from looker_sdk.sdk.api40 import methods as methods40  # type: ignore


def get_looker40sdk():
    sdk_attr = {
        "auth": "value_account",
        "deserialize": "value_user",
        "serialize": "value_password",
        "transport": "value_password",
        "api_version": "value_password",
    }
    return methods40.Looker40SDK(**sdk_attr)


MockUser = namedtuple("MockUser", ["id"])
MockRole = namedtuple("MockRole", ["permission_set"])
MockPermissionSet = namedtuple("MockPermissionSet", ["permissions"])


@patch("castor_extractor.visualization.looker.api.sdk.methods40.Looker40SDK")
def test_has_admin_permissions(mock_sdk):
    # unknown user
    mock_sdk.return_value.me.return_value = MockUser(id=None)
    sdk = get_looker40sdk()
    assert not has_admin_permissions(sdk)

    # set a user
    mock_sdk.return_value.me.return_value = MockUser(id="1")

    # user without roles
    mock_sdk.return_value.user_roles.return_value = []
    assert not has_admin_permissions(sdk)

    # user with multiple roles, but no "administer" permission
    mock_sdk.return_value.user_roles.return_value = [
        MockRole(MockPermissionSet(["access_data", "create_custom_fields"])),
        MockRole(MockPermissionSet(["explore"])),
    ]
    assert not has_admin_permissions(sdk)

    # user with "administer" permission
    mock_sdk.return_value.user_roles.return_value = [
        MockRole(MockPermissionSet(["access_data", "create_custom_fields"])),
        MockRole(MockPermissionSet(["explore", "administer"])),
    ]
    assert has_admin_permissions(sdk)



import logging
from argparse import ArgumentParser

from castor_extractor.visualization import metabase  # type: ignore

logging.basicConfig(level=logging.INFO, format="%(levelname)s - %(message)s")


def main():
    parser = ArgumentParser()

    parser.add_argument("-b", "--base-url", help="Metabase base url")
    parser.add_argument("-u", "--username", help="Metabase username")
    parser.add_argument("-p", "--password", help="Metabase password")

    parser.add_argument("-o", "--output", help="Directory to write to")

    args = parser.parse_args()

    client = metabase.ApiClient(
        base_url=args.base_url,
        user=args.username,
        password=args.password,
    )

    metabase.extract_all(
        client,
        output_directory=args.output,
    )



import os
import unittest
from unittest import mock

import requests_mock  # type: ignore
import tableauserverclient as TSC  # type: ignore

from ....client import ApiClient
from ..utils import KEYS

TEST_ASSET_DIR = os.path.join(os.path.dirname(__file__), "../assets/rest_api/")

GET_XML = os.path.join(TEST_ASSET_DIR, "workbook_get.xml")

EXPECTED = [
    {
        "id": "6d13b0ca-043d-4d42-8c9d-3f3313ea3a00",
        "name": "Superstore",
        "description": "description for Superstore",
        "tags": set(),
        "project_id": "ee8c6e70-43b6-11e6-af4f-f7b0d8e20760",
        "created_at": None,
        "updated_at": None,
        "owner_id": "5de011f8-5aa9-4d5b-b991-f462c8dd6bb7",
        "webpage_url": "http://tableauserver/#/workbooks/1/views",
    },
    {
        "id": "3cc6cd06-89ce-4fdc-b935-5294135d6d42",
        "name": "SafariSample",
        "description": "description for SafariSample",
        "tags": {"Sample", "Safari"},
        "project_id": "ee8c6e70-43b6-11e6-af4f-f7b0d8e20760",
        "created_at": None,
        "updated_at": None,
        "owner_id": "5de011f8-5aa9-4d5b-b991-f462c8dd6bb7",
        "webpage_url": "http://tableauserver/#/workbooks/2/views",
    },
]


class WorkbookTests(unittest.TestCase):
    @mock.patch.dict(os.environ, KEYS)
    def setUp(self):
        self._client = ApiClient(test=True)
        self._client._server = TSC.Server("http://test")

        # Fake signin
        self._client._server._site_id = "dad65087-b08b-4603-af4e-2887b8aafc67"
        self._client._server._auth_token = "j80k54ll2lfMZ0tv97mlPvvSCRyD0DOM"

        self.baseurl = self._client._server.workbooks.baseurl

    def test_fetch_workbooks(self):
        with open(GET_XML, "rb") as f:
            response_xml = f.read().decode("utf-8")
        with requests_mock.mock() as m:
            m.get(self.baseurl, text=response_xml)
            results = self._client._fetch_workbooks()

            self.assertEqual(results, EXPECTED)



import logging
from argparse import ArgumentParser

from castor_extractor.visualization import salesforce_reporting  # type: ignore

logging.basicConfig(level=logging.INFO, format="%(levelname)s - %(message)s")


def main():
    parser = ArgumentParser()

    parser.add_argument("-u", "--username", help="Salesforce username")
    parser.add_argument("-p", "--password", help="Salesforce password")
    parser.add_argument("-k", "--consumer-key", help="Salesforce consumer key")
    parser.add_argument(
        "-s", "--consumer-secret", help="Salesforce consumer secret"
    )
    parser.add_argument(
        "-t", "--security-token", help="Salesforce security token"
    )
    parser.add_argument("-l", "--url", help="Salesforce instance URL")
    parser.add_argument("-o", "--output", help="Directory to write to")

    args = parser.parse_args()
    salesforce_reporting.extract_all(
        username=args.username,
        password=args.password,
        consumer_key=args.consumer_key,
        consumer_secret=args.consumer_secret,
        security_token=args.security_token,
        instance_url=args.url,
        output_directory=args.output,
    )



import json
import logging
import os
from typing import Optional

from sqlalchemy import create_engine
from sqlalchemy.engine import Engine
from sqlalchemy.exc import OperationalError

from .....utils import (
    ExtractionQuery,
    PostgresClient,
    SerializedAsset,
    from_env,
)
from ...assets import EXPORTED_FIELDS, MetabaseAsset
from ...errors import EncryptionSecretKeyRequired, MetabaseLoginError
from ..decryption import decrypt
from ..shared import DETAILS_KEY, get_dbname_from_details
from .credentials import CredentialsDb, CredentialsDbKey, get_value

logger = logging.getLogger(__name__)

CURRENT_DIR = os.path.dirname(os.path.abspath(__file__))

SQL_FILE_PATH = "queries/{name}.sql"

ENCRYPTION_SECRET_KEY = "CASTOR_METABASE_ENCRYPTION_SECRET_KEY"  # noqa: S105
REQUIRE_SSL_KEY = "CASTOR_METABASE_REQUIRE_SSL_KEY"  # noqa: S105


def _optional_arg(args: dict, client_key: str, env_key: str) -> Optional[str]:
    arg_value = args.get(client_key)
    if arg_value is not None:
        return arg_value
    return from_env(env_key, allow_missing=True)


class DbClient(PostgresClient):
    """
    Connect to Metabase Database and fetch main assets.
    """

    @staticmethod
    def name() -> str:
        """return the name of the client"""
        return "Metabase/DB"

    def _engine_options(self, credentials: dict) -> dict:
        sslmode = "require" if self.require_ssl else "prefer"
        return {"connect_args": {"sslmode": sslmode}}

    def __init__(
        self,
        **kwargs,
    ):
        self._credentials = CredentialsDb.from_env(kwargs)
        self.encryption_secret_key = _optional_arg(
            kwargs, "encryption_secret_key", ENCRYPTION_SECRET_KEY
        )
        self.require_ssl = bool(
            _optional_arg(kwargs, "require_ssl", REQUIRE_SSL_KEY)
        )
        try:
            super().__init__(self._credentials.to_dict())
        except OperationalError as err:
            raise MetabaseLoginError(
                credentials_info=self._credentials.to_dict(hide=True),
                error_details=err.args,
            )

    def _load_query(self, name: str) -> ExtractionQuery:
        """load SQL text from file"""
        filename = SQL_FILE_PATH.format(name=name)
        path = os.path.join(CURRENT_DIR, filename)
        with open(path, "r") as f:
            content = f.read()
            statement = content.format(schema=self._credentials.schema)
            return ExtractionQuery(statement=statement, params={})

    def base_url(self) -> str:
        """Fetches the `base_url` of the Metabase instance"""
        query = self._load_query(name="base_url")
        rows = list(self.execute(query))
        return rows[0]["value"]

    def _database_specifics(
        self,
        databases: SerializedAsset,
    ) -> SerializedAsset:
        for db in databases:
            assert DETAILS_KEY in db  # this field is expected in database table

            try:
                details = json.loads(db[DETAILS_KEY])
            except json.decoder.JSONDecodeError as err:
                if not self.encryption_secret_key:
                    raise EncryptionSecretKeyRequired(
                        credentials_info=self._credentials.to_dict(hide=True),
                        error_details=err.args,
                    )
                decrypted = decrypt(db[DETAILS_KEY], self.encryption_secret_key)
                details = json.loads(decrypted)

            db["dbname"] = get_dbname_from_details(details)

        return databases

    def fetch(self, asset: MetabaseAsset) -> SerializedAsset:
        """fetches the given asset"""
        query = self._load_query(asset.value.lower())
        assets = list(self.execute(query))

        if asset == MetabaseAsset.DATABASE:
            assets = self._database_specifics(assets)

        logger.info(f"Fetching {asset.name} ({len(assets)} results)")

        # keep interesting fields
        return [
            {key: e.get(key) for key in EXPORTED_FIELDS[asset]} for e in assets
        ]



from unittest.mock import patch

import pytest

from .client import DEFAULT_PORT, MySQLClient, _check_required_keys


def test__check_required_keys():
    invalid_creds = {"user": "toto"}
    with pytest.raises(KeyError):
        _check_required_keys(invalid_creds)

    valid_creds = {"user": "toto", "password": "toto", "host": "localhost"}
    _check_required_keys(valid_creds)

    valid_creds_with_other_keys = {**valid_creds, "some": "other", "keys": 0}
    _check_required_keys(valid_creds_with_other_keys)


@patch.object(MySQLClient, "__init__")
def test_build_uri(mocked_client):
    credentials = {
        "user": "peanut",
        "password": "butter",
        "host": "jelly",
    }
    result = MySQLClient._build_uri(mocked_client, credentials)
    assert result == f"mysql+pymysql://peanut:butter@jelly:{DEFAULT_PORT}"

    port = 4567
    credentials_with_port = {**credentials, "port": port}
    result = MySQLClient._build_uri(mocked_client, credentials_with_port)
    assert result == f"mysql+pymysql://peanut:butter@jelly:{port}"



import logging
from typing import Iterable, Optional, Tuple, Union

from ...utils import (
    OUTPUT_DIR,
    current_timestamp,
    deep_serialize,
    from_env,
    get_output_filename,
    write_json,
    write_summary,
)
from .assets import SalesforceReportingAsset
from .client import SalesforceClient, SalesforceCredentials

logger = logging.getLogger(__name__)


def iterate_all_data(
    client: SalesforceClient,
) -> Iterable[Tuple[str, Union[list, dict]]]:
    """Iterate over the extracted data from Salesforce"""

    for asset in SalesforceReportingAsset:
        logger.info(f"Extracting {asset.value.upper()} from REST API")
        data = client.fetch(asset)
        yield asset.name.lower(), deep_serialize(data)


def extract_all(
    username: str,
    password: str,
    consumer_key: str,
    consumer_secret: str,
    security_token: str,
    instance_url: str,
    output_directory: Optional[str] = None,
) -> None:
    """
    Extract data from Salesforce REST API
    Store the output files locally under the given output_directory
    """
    _output_directory = output_directory or from_env(OUTPUT_DIR)
    creds = SalesforceCredentials(
        username=username,
        password=password,
        consumer_key=consumer_key,
        consumer_secret=consumer_secret,
        security_token=security_token,
    )
    client = SalesforceClient(credentials=creds, instance_url=instance_url)
    ts = current_timestamp()

    for key, data in iterate_all_data(client):
        filename = get_output_filename(key, _output_directory, ts)
        write_json(filename, data)

    write_summary(_output_directory, ts)



import logging
from typing import Any, Callable, Dict, Iterator, List, Optional, Tuple

import requests

from ..assets import SalesforceReportingAsset
from .constants import DEFAULT_API_VERSION, DEFAULT_PAGINATION_LIMIT
from .credentials import SalesforceCredentials
from .soql import queries

logger = logging.getLogger(__name__)


class SalesforceClient:
    """
    Salesforce API client.
    https://developer.salesforce.com/docs/atlas.en-us.api_rest.meta/api_rest/intro_rest.htm
    """

    api_version = DEFAULT_API_VERSION
    pagination_limit = DEFAULT_PAGINATION_LIMIT

    def __init__(
        self,
        credentials: SalesforceCredentials,
        instance_url: Optional[str] = None,
    ):
        self.credentials = credentials
        self.instance_url = instance_url
        self._token = self._access_token()

    def _access_token(self) -> Tuple[str, str]:
        url = f"{self.instance_url}/services/oauth2/token"
        response = self._call(
            url, "POST", data=self.credentials.token_request_payload()
        )
        return response["access_token"]

    def _header(self) -> Dict:
        return {"Authorization": f"Bearer {self._token}"}

    @staticmethod
    def _call(
        url: str,
        method: str = "GET",
        *,
        header: Optional[Dict] = None,
        params: Optional[Dict] = None,
        data: Optional[Dict] = None,
        processor: Optional[Callable] = None,
    ) -> Any:
        logger.debug(f"Calling {method} on {url}")
        result = requests.request(
            method,
            url,
            headers=header,
            params=params,
            data=data,
        )
        result.raise_for_status()

        if processor:
            return processor(result)

        return result.json()

    def _query(
        self, url: str, params: Optional[Dict] = None
    ) -> Tuple[List[Dict], Optional[str]]:
        results = self._call(url, params=params, header=self._header())
        return results["records"], results.get("nextRecordsUrl")

    def _query_all(self, query: str) -> Iterator[Dict]:
        """
        Run a SOQL query over salesforce API.

        more: https://developer.salesforce.com/docs/atlas.en-us.api_rest.meta/api_rest/dome_query.htm
        """
        url = f"{self.instance_url}/services/data/v{self.api_version}/query"
        records, next_page = self._query(url, params={"q": query})
        yield from records

        page_count = 0
        while next_page or page_count >= self.pagination_limit:
            logger.info(f"querying page {page_count}")
            url = f"{self.instance_url}{next_page}"
            records, next_page = self._query(url)
            yield from records
            page_count += 1

    def fetch(self, asset: SalesforceReportingAsset) -> List[Dict]:
        logger.info(f"Starting extraction of {asset}")
        return list(self._query_all(queries[asset]))



import logging

from ...utils import LocalStorage, from_env, write_summary
from ..abstract import (
    CATALOG_ASSETS,
    EXTERNAL_LINEAGE_ASSETS,
    QUERIES_ASSETS,
    VIEWS_ASSETS,
    ExtractionProcessor,
    SupportedAssets,
    WarehouseAsset,
    WarehouseAssetGroup,
    common_args,
    extractable_asset_groups,
)
from .client import MySQLClient
from .query import MySQLQueryBuilder

logger = logging.getLogger(__name__)


MYSQL_ASSETS: SupportedAssets = {
    WarehouseAssetGroup.CATALOG: CATALOG_ASSETS,
    WarehouseAssetGroup.QUERY: QUERIES_ASSETS,
    WarehouseAssetGroup.VIEW_DDL: VIEWS_ASSETS,
    WarehouseAssetGroup.ROLE: (WarehouseAsset.USER,),
    WarehouseAssetGroup.EXTERNAL_LINEAGE: EXTERNAL_LINEAGE_ASSETS,
}

MYSQL_USER = "CASTOR_MYSQL_USER"
MYSQL_PASSWORD = "CASTOR_MYSQL_PASSWORD"  # noqa: S105
MYSQL_HOST = "CASTOR_MYSQL_HOST"
MYSQL_PORT = "CASTOR_MYSQL_PORT"


def _credentials(params: dict) -> dict:
    """Extract MySQL credentials"""

    return {
        "user": params.get("user") or from_env(MYSQL_USER),
        "password": params.get("password") or from_env(MYSQL_PASSWORD),
        "host": params.get("host") or from_env(MYSQL_HOST),
        "port": params.get("port") or from_env(MYSQL_PORT, allow_missing=True),
    }


def extract_all(**kwargs) -> None:
    """
    Extract all assets from MySQL and store the results in CSV files
    """
    output_directory, skip_existing = common_args(kwargs)

    client = MySQLClient(credentials=_credentials(kwargs))
    query_builder = MySQLQueryBuilder()

    storage = LocalStorage(directory=output_directory)
    extractor = ExtractionProcessor(
        client=client,
        query_builder=query_builder,
        storage=storage,
    )

    for group in extractable_asset_groups(MYSQL_ASSETS):
        for asset in group:
            logger.info(f"Extracting `{asset.value.upper()}` ...")
            location = extractor.extract(asset, skip_existing)
            logger.info(f"Results stored to {location}\n")

    write_summary(
        output_directory,
        storage.stored_at_ts,
        client_name=client.name(),
    )



import logging
from typing import Dict, Iterator, List, cast

import requests
from requests import HTTPError, Response

from .....utils import JsonType, SerializedAsset
from ...assets import EXPORTED_FIELDS, MetabaseAsset
from ...errors import MetabaseLoginError, SuperuserCredentialsRequired
from ...types import IdsType
from ..shared import DETAILS_KEY, get_dbname_from_details
from .credentials import CredentialsApi, CredentialsApiKey, get_value

logger = logging.getLogger(__name__)

URL_TEMPLATE = "{base_url}/api/{endpoint}"

ROOT_KEY = "root"
CARDS_KEY = "dashcards"
CARDS_KEY_DEPRECATED = "ordered_cards"  # prior to v0.48 of Metabase
DATA_KEY = "data"


class ApiClient:
    """
    Connect to Metabase API and fetch main assets.
    Superuser credentials are required.
    https://www.metabase.com/docs/latest/api-documentation.html
    """

    def __init__(
        self,
        **kwargs,
    ):
        self._credentials = CredentialsApi(
            base_url=get_value(CredentialsApiKey.BASE_URL, kwargs),
            user=get_value(CredentialsApiKey.USER, kwargs),
            password=get_value(CredentialsApiKey.PASSWORD, kwargs),
        )
        self._session = requests.Session()
        self._session_id = self._login()
        self._check_permissions()  # verify that the given user is superuser

    @staticmethod
    def name() -> str:
        """return the name of the client"""
        return "Metabase/API"

    def base_url(self) -> str:
        """Return base_url from credentials"""
        return self._credentials.base_url

    def _url(self, endpoint: str) -> str:
        return URL_TEMPLATE.format(
            base_url=self._credentials.base_url,
            endpoint=endpoint,
        )

    def _headers(self) -> dict:
        return {
            "X-Metabase-Session": self._session_id,
            "Content-type": "application/x-www-form-urlencoded",
        }

    @staticmethod
    def _answer(response: Response):
        answer = response.json()
        if isinstance(answer, Dict) and DATA_KEY in answer:
            # v0.41 of Metabase introduced embedded data for certain calls
            # {'data': [{ }, ...] , 'total': 15, 'limit': None, 'offset': None}"
            return answer[DATA_KEY]
        return answer

    def _call(self, endpoint: str) -> JsonType:
        url = self._url(endpoint)
        headers = self._headers()
        response = self._session.get(url=url, headers=headers)
        response.raise_for_status()  # check for errors
        return self._answer(response)

    def _check_permissions(self) -> None:
        try:
            # This endpoint requires superuser credentials
            self._call("collection/graph")
        except HTTPError as err:
            if err.response.status_code == 403:  # forbidden
                raise SuperuserCredentialsRequired(
                    credentials_info=self._credentials.to_dict(hide=True),
                    error_details=err.args,
                )
            raise

    def _login(self) -> str:
        url = self._url("session")
        payload = {
            "username": self._credentials.user,
            "password": self._credentials.password,
        }
        response = self._session.post(url, json=payload)
        logger.info(f"Getting session_id: {response.json()}")

        if not response.json().get("id"):
            raise MetabaseLoginError(
                credentials_info=self._credentials.to_dict(hide=True),
                error_details=response.json(),
            )

        return response.json()["id"]

    def _fetch_ids(self, asset: MetabaseAsset) -> IdsType:
        ids: IdsType = []
        results = self._call(endpoint=asset.name.lower())
        for res in cast(List, results):
            assert isinstance(res, dict)
            ids.append(res["id"])
        return ids

    def _fetch_details(self, asset: MetabaseAsset) -> SerializedAsset:
        """Fetches list of IDS first, then iterate on api/asset/:id to fetch details"""
        result: SerializedAsset = []
        for id_ in self._fetch_ids(asset):
            details = self._call(endpoint=f"{asset.name.lower()}/{id_}")
            result.append(cast(Dict, details))
        return result

    @staticmethod
    def _collection_specifics(collections: SerializedAsset) -> SerializedAsset:
        # remove the root folder
        def _is_not_root(collection: dict) -> bool:
            return collection.get("id") != ROOT_KEY

        return list(filter(_is_not_root, collections))

    @staticmethod
    def _database_specifics(databases: SerializedAsset) -> SerializedAsset:
        for db in databases:
            # superuser privileges are mandatory, this field should be present
            assert DETAILS_KEY in db
            details = db[DETAILS_KEY]
            db["dbname"] = get_dbname_from_details(details)

        return databases

    @staticmethod
    def _dashboard_cards(dashboards: SerializedAsset) -> Iterator[Dict]:
        for d in dashboards:
            d_cards = d.get(CARDS_KEY) or d.get(CARDS_KEY_DEPRECATED) or []
            yield from d_cards

    def fetch(self, asset: MetabaseAsset) -> SerializedAsset:
        """fetches the given asset"""
        if asset == MetabaseAsset.DASHBOARD_CARDS:
            dashboards = self._fetch_details(MetabaseAsset.DASHBOARD)
            assets = list(self._dashboard_cards(dashboards))
        else:
            answer = self._call(asset.name.lower())
            assets = cast(List, answer)

        if asset == MetabaseAsset.DATABASE:
            assets = self._database_specifics(assets)

        if asset == MetabaseAsset.COLLECTION:
            assets = self._collection_specifics(assets)

        logger.info(f"Fetching {asset.name} ({len(assets)} results)")

        # keep interesting fields
        return [
            {key: e.get(key) for key in EXPORTED_FIELDS[asset]} for e in assets
        ]



from enum import Enum

EXTRACTION_BUCKET = "extraction-storage"


class FileType(Enum):
    """type of file to load"""

    DBT = "DBT"
    QUALITY = "QUALITY"
    VIZ = "VIZ"
    WAREHOUSE = "WAREHOUSE"


PATH_TEMPLATES = {
    FileType.DBT: "transformation-{source_id}/{timestamp}-manifest.json",
    FileType.QUALITY: "quality-{source_id}/{timestamp}-{filename}",
    FileType.VIZ: "visualization-{source_id}/{filename}",
    FileType.WAREHOUSE: "warehouse-{source_id}/{filename}",
}


"""
The default request timeout in seconds for the upload
"""
DEFAULT_TIMEOUT = 60.0
ENVIRON_TIMEOUT = "CASTOR_TIMEOUT_OVERRIDE"

"""
The default retry for the upload
"""
DEFAULT_RETRY = 1
ENVIRON_RETRY = "CASTOR_RETRY_OVERRIDE"



import logging
from argparse import ArgumentParser

from castor_extractor.visualization import qlik  # type: ignore

logging.basicConfig(level=logging.INFO, format="%(levelname)s - %(message)s")


def main():
    parser = ArgumentParser()

    parser.add_argument("-b", "--base-url", help="Qlik base url")
    parser.add_argument("-a", "--api-key", help="Generated API key")
    parser.add_argument("-o", "--output", help="Directory to write to")
    parser.add_argument(
        "-e",
        "--except-http-error-statuses",
        type=int,
        nargs="+",
        help="List of HTTP statuses for which to catch errors from and log "
        "as warning instead. Helpful to continue script execution when "
        "missing rights on some assets.",
    )

    args = parser.parse_args()
    qlik.extract_all(
        base_url=args.base_url,
        api_key=args.api_key,
        output_directory=args.output,
        except_http_error_statuses=args.except_http_error_statuses,
    )



from dataclasses import dataclass
from enum import Enum
from typing import Dict

from ....utils import from_env


class CredentialsKey(Enum):
    """keys present in dbt credentials"""

    TOKEN = "token"  # noqa: S105
    JOB_ID = "job_id"


CREDENTIALS_ENV: Dict[CredentialsKey, str] = {
    CredentialsKey.TOKEN: "CASTOR_DBT_TOKEN",
    CredentialsKey.JOB_ID: "CASTOR_DBT_JOB_ID",
}


def get_value(key: CredentialsKey, kwargs: dict) -> str:
    """
    Returns the value of the given key:
    - from kwargs in priority
    - from ENV if not provided (raises an error if not found in ENV)
    """
    env_key = CREDENTIALS_ENV[key]
    return str(kwargs.get(key.value) or from_env(env_key))


@dataclass
class DbtCredentials:
    token: str
    job_id: str



from datetime import date, datetime
from enum import Enum
from typing import Any, Iterator, Union, overload
from uuid import UUID

from .type import Getter


@overload
def deep_serialize(__value: None) -> None: ...


@overload
def deep_serialize(__value: int) -> int: ...


@overload
def deep_serialize(__value: float) -> float: ...


@overload
def deep_serialize(__value: Union[str, date, datetime, Enum]) -> str: ...


@overload
def deep_serialize(__value: Union[list, tuple]) -> list: ...


@overload
def deep_serialize(__value: dict) -> dict: ...


@overload
def deep_serialize(__value: Iterator) -> Iterator: ...


def deep_serialize(value: Any) -> Any:
    """Deep serialize any data to primitive"""
    if isinstance(value, (float, int, str)) or value is None:
        return value

    if isinstance(value, (date, datetime)):
        return value.isoformat()

    if isinstance(value, UUID):
        return str(value)

    if isinstance(value, Enum):
        return value.value

    if isinstance(value, (tuple, list, set)):
        return [deep_serialize(el) for el in value]

    if isinstance(value, dict):
        return {k: deep_serialize(v) for k, v in value.items()}

    if isinstance(value, Iterator):
        return (deep_serialize(item) for item in value)

    try:
        items = dict(value).items()
        return {k: deep_serialize(v) for k, v in items}
    except TypeError:
        raise ValueError(f"Value {str(value)} is not serializable")


def getproperty(element: Any, getter: Getter) -> Any:
    """
    Access a property for an instance or a dict
    Access an index for a list or a tuple
    Either a callback, a string or an int can be given
    """

    if isinstance(getter, str):
        if isinstance(element, dict):
            return element[getter]
        return getattr(element, getter)
    if isinstance(getter, int) and isinstance(element, (tuple, list)):
        return element[getter]
    return getter(element)



import logging
from argparse import ArgumentParser

from castor_extractor.warehouse import postgres  # type:ignore

logging.basicConfig(level=logging.INFO, format="%(levelname)s - %(message)s")


def main():
    parser = ArgumentParser()

    parser.add_argument("-H", "--host", help="Postgres Host")
    parser.add_argument("-P", "--port", help="Postgres Port")
    parser.add_argument("-d", "--database", help="Postgres Database")
    parser.add_argument("-u", "--user", help="Postgres User")
    parser.add_argument("-p", "--password", help="Postgres Password")

    parser.add_argument("-o", "--output", help="Directory to write to")

    parser.add_argument(
        "--skip-existing",
        dest="skip_existing",
        action="store_true",
        help="Skips files already extracted instead of replacing them",
    )
    parser.set_defaults(skip_existing=False)

    args = parser.parse_args()

    postgres.extract_all(
        host=args.host,
        database=args.database,
        port=args.port,
        user=args.user,
        password=args.password,
        output_directory=args.output,
        skip_existing=args.skip_existing,
    )



import pytest

from .credentials import SnowflakeCredentials


def test_SnowflakeCredentials():
    assert SnowflakeCredentials(
        account="snowwhite",
        password="hammer",
        user="dwarf",
    )

    assert SnowflakeCredentials(
        account="snowwhite",
        private_key="pickaxe",
        user="dwarf",
    )

    with pytest.raises(ValueError):
        SnowflakeCredentials(
            account="snowwhite",
            password="hammer",
            private_key="pickaxe",
            user="dwarf",
        )

    with pytest.raises(ValueError):
        SnowflakeCredentials(
            account="snowwhite",
            user="dwarf",
        )



from enum import Enum


class Issue(Enum):
    """kind of issues we might detect while checking files"""

    MISSING_VALUE = "missing_value"
    UNAUTHORIZED_VALUE = "unauthorized_value"
    WRONG_DATATYPE = "wrong_datatype"
    UNKNOWN_REFERENCE = "unknown_reference"
    DUPLICATE_VALUE = "duplicate_value"
    REPEATED_QUOTES = "repeated_quotes"


class DataType(Enum):
    """data types supported by the file checker"""

    STRING = "string"
    FLOAT = "float"
    INTEGER = "integer"
    DATETIME = "datetime"
    LIST = "list"



from .env import _parse_float, _parse_int


def test__parse_float():
    value = _parse_float("66", 33.0)
    assert value == 66.0
    assert isinstance(value, float)

    value = _parse_float(None, 66.0)
    assert value == 66.0
    assert isinstance(value, float)


def test__parse_int():
    value = _parse_int("66", 33)
    assert value == 66
    assert isinstance(value, int)

    value = _parse_int(None, 33)
    assert value == 33
    assert isinstance(value, int)



from typing import Iterable, List, Optional, Tuple, Union

from ...utils import (
    OUTPUT_DIR,
    current_timestamp,
    deep_serialize,
    from_env,
    get_output_filename,
    write_json,
    write_summary,
)
from .assets import METADATA_ASSETS, PowerBiAsset
from .client import Client, Credentials


def iterate_all_data(
    client: Client,
) -> Iterable[Tuple[PowerBiAsset, Union[List, dict]]]:
    for asset in PowerBiAsset:
        if asset in METADATA_ASSETS:
            continue

        data = client.fetch(asset)
        yield asset, deep_serialize(data)


def extract_all(
    tenant_id: str,
    client_id: str,
    secret: str,
    scopes: Optional[List[str]] = None,
    output_directory: Optional[str] = None,
) -> None:
    """
    Extract data from PowerBI REST API
    Store the output files locally under the given output_directory
    """
    _output_directory = output_directory or from_env(OUTPUT_DIR)
    creds = Credentials(
        tenant_id=tenant_id,
        client_id=client_id,
        secret=secret,
        scopes=scopes,
    )
    client = Client(creds)
    ts = current_timestamp()

    for key, data in iterate_all_data(client):
        filename = get_output_filename(key.name.lower(), _output_directory, ts)
        write_json(filename, data)

    write_summary(_output_directory, ts)



from datetime import date, datetime, timedelta

from .client import FORMAT, _at_midnight, _is_expired

NEVER_EXPIRING_DATE = datetime.max
TODAY_DATE = _at_midnight(date.today())
EXPIRED_DATE = _at_midnight(date.today() - timedelta(days=2))


def test__is_expired():
    test_datetimes = (
        ({"time": NEVER_EXPIRING_DATE.strftime(FORMAT)}, False),
        ({"time": TODAY_DATE.strftime(FORMAT)}, False),
        ({"time": EXPIRED_DATE.strftime(FORMAT)}, True),
    )

    for test_datetime, is_expired in test_datetimes:
        assert _is_expired(test_datetime) == is_expired
